import {HttpMethod, Route, StatusCodes} from 'org.crazydoctor.expressts';
import {Request, Response} from 'express';
import { ClassMapEntity, Sources } from '../sources/Sources';
import fs from 'fs';
import ServerApp from '..';
import { ISession } from '../session/ISession';
import { CommentsManager } from '../comments/CommentsManager';

class GetClass extends Route {
	private processClass(cls: ClassMapEntity): { Class: ClassMapEntity, ClassSource: string } {
		const srcPath = cls.filePath;
		const source = fs.readFileSync(`${Sources.sourcesCopyPath}/${srcPath}`, { encoding: 'utf8' }).toString();
		return { Class: cls, ClassSource: source };
	}

	protected action = (req: Request, res: Response): any => {
		const session = req.session as ISession;
		
		if(ServerApp.SourcesUpdating) {
			res.status(StatusCodes.ACCEPTED).render('sources-update');
			return;
		}

		const className = req.params.className;
		const cls: ClassMapEntity | null = Sources.findClass(className);
		if(!cls) {
			res.status(StatusCodes.NOT_FOUND).render('class', { isAdmin: session.isAdmin || false, isEditor: session.isEditor || false, Login: session.login || null, Class: className, ClassSource: '', template: 'class', title: 'Class not found', ClassList: Sources.getShortenedClassMap(), RepoNames: Sources.getRepoNames(), Z8Locales: null, Comments: null });
		} else {
			CommentsManager.getCommentsByClass(cls.root, cls.name).then((result) => {
				if(!result.success) {
					res.status(StatusCodes.INTERNAL_SERVER_ERROR).send('Internal server error');
					return;
				}

				const commentsObj: {[key: string] : any} = {};
				result.comments?.forEach((comment) => {
					commentsObj[comment.propertyName] = comment;
				});
				res.status(StatusCodes.OK).render('class', Object.assign(this.processClass(cls), { isAdmin: session.isAdmin || false, isEditor: session.isEditor || false, Login: session.login || null, template: 'class', title: `Class: ${className}`, ClassList: Sources.getShortenedClassMap(), RepoNames: Sources.getRepoNames(), Z8Locales: Sources.Z8Locales, Comments: commentsObj }));
			});
		}
	};
	protected method = HttpMethod.GET;
	protected order = 1;
	protected route = '/class/:className';
}

export default GetClass;