import { HttpMethod, Route, StatusCodes } from 'org.crazydoctor.expressts';
import { Request, Response } from 'express';
import { ISession } from '../session/ISession';
import { CommentUpdateStatus, CommentsManager } from '../comments/CommentsManager';
import ServerApp from '..';

class PostUpdateComment extends Route {
	protected action = (req: Request, res: Response): any => {
		const session = req.session as ISession;
		const params = req.body;

		const root = params.root.trim();
		const className = params.class.trim();
		const propertyName = params.property.trim();
		const comment = params.comment;

		if(!session.isEditor) {
			res.status(StatusCodes.FORBIDDEN).send('Access denied');
			return;
		}

		if(ServerApp.SourcesUpdating) {
			res.status(StatusCodes.FORBIDDEN).send('Sources are being updated');
			return;
		}

		CommentsManager.update({
			root: root,
			className: className,
			propertyName: propertyName,
			text: comment,
			timestamp: new Date().getTime(),
			author: session.login!
		});

		res.status(StatusCodes.ACCEPTED).send('ENQUEUED');
	};

	protected method = HttpMethod.POST;
	protected order = 5;
	protected route = '/updateComment';
}

export default PostUpdateComment;